from .client import Client
from .packet import Packet
from . import utils

from .__version__ import __author__, __title__, __description__
from .__version__ import __url__, __version__, __credits__, __license__


import asyncio

from .utils import Packet

class Socket:
	def __init__(self, host, port, loop=None):
		self.loop = loop or asyncio.get_event_loop()

		self.__socket = asyncio.open_connection(host, port, loop=self.loop)
		self._reader:asyncio.StreamReader = None
		self._writer:asyncio.StreamWrite = None
		self.connected = False

	async def connect(self):
		"""Connect the socket to the host."""
		self._reader, self._writer = await self.__socket
		self.connected = True

		del self.__socket

	async def recv(self, size):
		"""Receive up to :size: bytes from the socket."""
		return await self._reader.read(size)

	async def send(self, data):
		"""Send a data string to the socket."""
		rval = self._writer.write(data)
		await self.flush()
		return rval

	async def flush(self):
		"""Flush send buffer."""
		await self._writer.drain()

	def close(self):
		"""Close the socket."""
		self.connected = False
		self._writer.close()

class Connection:
	def __init__(self, name, client, loop=None):
		self.name = name
		self.client = client
		self.loop = loop

		self.socket = None
		self.address = ()
		self.fingerprint = 0

		self.open = False

	async def connect(self, host, port):
		"""coro
		Connect the client to the server host:port"""
		self.address = (host, port)
		self.socket = Socket(host, port, self.loop)

		await self.socket.connect()
		self.open = True

		self.client.dispatch('connection_made', self)

		asyncio.ensure_future(self._recv_loop())

	async def _recv_loop(self):
		while self.open:
			lensize = await self.socket.recv(1)
			if len(lensize)==0:
				self.close()
				if self.open:
					raise EOFError('The connection "{.name}" has been closed.'.format(self))
				break
			length = int.from_bytes(await self.socket.recv(lensize[0]), 'big')
			data = await self.socket.recv(length)
			await self.client.receive_packet(Packet(data), self)

	async def send(self, packet):
		"""coro
		Send a packet to the socket"""

		# print('sending', packet.export(self.fingerprint))

		await self.socket.send(packet.export(self.fingerprint))
		self.fingerprint = (self.fingerprint + 1) % 100

	async def close(self): # nothing is asynchronous but idc
		self.open = False
		self.socket.close()


import asyncio

from .packet import Packet
from .get_keys import get_keys
from .connection import Connection


class Client:
	LOG_UNHANDLED_PACKETS = False

	def __init__(self, loop=None):
		self.loop = loop or asyncio.get_event_loop()

		self.main = Connection('main', self, self.loop)
		self.bulle = None

		self.community = 0 # EN
		self.whisper_id = 0

	async def handle_packet(self, connection:Connection, packet:Packet):
		CCC = packet.readCode()
		if CCC==(26, 3): # Handshake ok
			online_players = packet.read32() # online players
			connection.fingerprint = packet.read8()
			community = packet.readString() # community
			country = packet.readString() # country
			self.authkey = packet.read32()

			self.loop.create_task(self._heartbeat_loop())

			await connection.send(Packet.new(8,2).write8(self.community).write8(0))

			os_info = Packet.new(28,17).writeString('en').writeString('Linux')
			os_info.writeString('LNX 29,0,0,140').write8(0)

			await connection.send(os_info)
			self.dispatch('login_ready', online_players, community, country)

		elif CCC==(5, 21): # Joined room
			private = packet.readBool()
			room_name = packet.readString() # Decode it at your own risk
			self.dispatch('joined_room', room_name, private)

		elif CCC==(16, 2): # Tribe invitation received
			author = packet.readUTF()
			tribe = packet.readUTF()
			self.dispatch('tribe_inv', author, tribe)

		elif CCC==(6, 6): # Room message
			player_id, username, commu, message = packet.unpack('LsBs')
			self.dispatch('room_message', username, message)

		elif CCC==(44, 1): # Bulle switching
			bulle_id = packet.read32()
			bulle_ip = packet.readString().decode()

			if self.bulle is not None:
				await self.bulle.close()

			self.bulle = Connection('bulle', self, self.loop)
			await self.bulle.connect(bulle_ip, self.main.address[1])
			await self.bulle.send(Packet.new(*CCC).write32(bulle_id))

		elif CCC==(44, 22): # Fingerprint offset changed
			connection.fingerprint = packet.read8()

		elif CCC==(60, 3): # Community platform
			TC = packet.read16()
			if TC==3: # Connected to the community platform
				self.dispatch('ready')
			elif TC==66: # Whisper
				author, commu, receiver, message = packet.readUTF(), packet.read32(), packet.readUTF(), packet.readUTF()
				self.dispatch('whisper', author, commu, receiver, message)
		else:
			if self.LOG_UNHANDLED_PACKETS:
				print(CCC, bytes(packet.buffer)[2:])

	async def receive_packet(self, packet, connection):
		self.dispatch('raw_socket', connection, packet.copy())
		await self.handle_packet(connection, packet)

	async def _heartbeat_loop(self):
		last_heartbeat = 0
		while self.main.open:
			if self.loop.time()-last_heartbeat>=10:
				await self.main.send(Packet.new(26, 26))
				if self.bulle is not None and self.bulle.open:
					await self.bulle.send(Packet.new(26, 26))

				last_heartbeat = self.loop.time()
			await asyncio.sleep(.5)

	async def start(self, api_tfmid, api_token):
		"""coro
		Starts the client."""

		self.keys = keys = await get_keys(api_tfmid, api_token)

		await self.main.connect('164.132.202.12', 5555)

		while not self.main.socket.connected:
			await asyncio.sleep(.1)

		packet = Packet.new(28, 1).write16(keys.version).writeString(keys.connection)
		packet.writeString('Desktop').writeString('-').write32(0x1fbd).writeString('')
		packet.writeString('74696720697320676f6e6e61206b696c6c206d7920626f742e20736f20736164')
		packet.writeString("A=t&SA=t&SV=t&EV=t&MP3=t&AE=t&VE=t&ACC=t&PR=t&SP=f&SB=f&DEB=f&V=LNX 29,0,0,140&M=Adobe Linux&R=1920x1080&COL=color&AR=1.0&OS=Linux&ARCH=x86&L=en&IME=t&PR32=t&PR64=t&LS=en-US&PT=Desktop&AVD=f&LFD=f&WD=f&TLS=t&ML=5.1&DP=72")
		packet.write32(0).write32(0x6257).writeString('')

		await self.main.send(packet)

	def event(self, coro):
		name = coro.__name__
		if not name.startswith('on_'):
			raise Exception("'{}' ins't a correct event naming.".format(name))
		if not asyncio.iscoroutinefunction(coro):
			raise Exception("Couldn't register a non-coroutine function for the event {}.".format(name))

		setattr(self, name, coro)
		return coro

	async def _run_event(self, coro, event_name, *args, **kwargs):
		try:
			await coro(*args, **kwargs)
		except asyncio.CancelledError:
			pass
		except Exception as e:
			if hasattr(self, 'on_error'):
				try:
					await self.on_error(event_name, e, *args, **kwargs)
				except asyncio.CancelledError:
					pass
				else:
					raise e
			else:
				raise e

	def dispatch(self, event, *args, **kwargs):
		method = 'on_' + event

		# add wait_for handling

		coro = getattr(self, method, None)
		if coro is not None:
			asyncio.ensure_future(self._run_event(coro, method, *args, **kwargs), loop=self.loop)

	async def login(self, username, password, room='1', encrypted=True):
		"""coro
		Log in the game."""

		if not encrypted:
			from .utils import shakikoo
			password = shakikoo(password)

		packet = Packet.new(26, 8).writeString(username).writeString(password)
		packet.writeString("app:/TransformiceAIR.swf/[[DYNAMIC]]/2/[[DYNAMIC]]/4")
		packet.writeString(room).write32(self.authkey^self.keys.auth)
		packet.cipher(self.keys.identification).write8(0)

		await self.main.send(packet)

	async def sendRoomMessage(self, message):
		"""coro
		Send a message to the room."""

		packet = Packet.new(6, 6).writeString(message)
		packet.xor_cipher(self.keys.msg, self.bulle.fingerprint)

		await self.bulle.send(packet)

	async def whisper(self, username, message):
		"""coro
		Whisper to a player."""

		self.whisper_id = (self.whisper_id + 1) % 0XFFFFFFFF

		packet = Packet.new(60, 3).write16(52).write32(self.whisper_id)
		packet.writeString(username).writeString(message)
		packet.xor_cipher(self.keys.msg, self.main.fingerprint)

		await self.main.send(packet)

	async def sendPrivateMessage(self, username, message):
		"""coro
		Deprecated alias for `Client.whisper`."""
		await self.whisper(username, message)

	async def playEmote(self, id, flag='be'):
		"""coro
		Play an emote id with it's optinnal flag (id = 10)."""

		packet = Packet.new(8, 1).write8(id).write32(0)
		if id==10:
			packet.writeString(flag)

		await self.bulle.send(packet)

	async def sendEmoticon(self, id):
		"""coro
		Send an emoticon to the server."""

		if 10>id>0:
			raise Exception('Invalid emoticon id')

		packet = Packet.new(8, 5).write8(id).write32(0)

		await self.bulle.send(packet)

	async def loadLua(self, lua_code):
		"""coro
		Load a lua code to the room."""

		if isinstance(lua_code, str):
			lua_code = lua_code.encode()

		packet = Packet.new(29, 1).write24(len(lua_code)).writeBytes(lua_code)

		await self.bulle.send(packet)

	async def sendCommand(self, cmd):
		"""coro
		Send a command to the game."""

		packet = Packet.new(6, 26).writeString(cmd)
		packet.xor_cipher(self.keys.msg, self.main.fingerprint)

		await self.main.send(packet)

	async def enterTribe(self):
		"""coro
		Enter the tribe house"""

		await self.main.send(Packet.new(16, 1))

	async def enterTribeHouse(self):
		"""coro
		Alias for Client.enterTribe"""

		await self.enterTribe()

	async def joinRoom(self, room_name):
		"""coro
		Join a room.
		The event 'on_joined_room' is dispatched when the client has successfully joined the room."""

		await self.main.send(Packet.new(5, 35).write8(255).writeString(room_name))

	async def enterInvTribeHouse(self, author):
		"""coro
		Join the tribe house of another player after receiving an /inv."""

		await self.main.send(Packet.new(16, 2).writeString(author))



class Packet:
	def __init__(self, buffer=None):
		if buffer is None:
			buffer = bytearray()
		elif not isinstance(buffer, bytearray):
			buffer = bytearray(buffer)

		self.buffer = buffer
		self.pos = 0

		self.exported = False

	def __repr__(self):
		return '<Packet {!r}>'.format(bytes(self.buffer))

	@classmethod
	def new(cls, c, cc=None):
		msg = cls()
		if cc is None:
			msg.write16(c)
		else:
			msg.write8(c).write8(cc)

		return msg

	def copy(self):
		return Packet(self.buffer[:])

	def unpack(self, fmt):
		results = []
		formats = {
			'B': self.read8,
			'H': self.read16,
			'I': self.read24,
			'L': self.read32,
			's': self.readString
		}
		times = 1
		for c in fmt:
			if c.isdigit():
				if times==1:
					times = int(c)
				else:
					times = times * 10 + int(c)
			else:
				for i in range(times):
					results.append(formats[c]())
				times = 1

		return results

	def readBytes(self, nbr=1):
		self.pos += nbr
		return self.buffer[self.pos-nbr:self.pos]

	def readCode(self):
		return self.read8(), self.read8()

	def read8(self):
		self.pos += 1
		try:
			return self.buffer[self.pos-1]
		except:
			return 0

	def read16(self):
		return (self.read8() << 8) + self.read8()

	def read24(self):
		return (self.read16() << 8) + self.read8()

	def read32(self):
		return (self.read24() << 8) + self.read8()

	def readBool(self):
		return self.read8()==1

	def readString(self):
		"""return a encoded string (in bytes)"""
		return bytes(self.readBytes(self.read16()))

	def readUTF(self):
		"""return a decoded string"""
		return self.readString().decode()

	def writeBytes(self, bytes):
		self.buffer.extend(bytes)
		return self

	def writeCode(self, c, cc):
		return self.write8(c).write8(cc)

	def write8(self, value):
		self.buffer.append(value&0xff)
		return self

	def write16(self, value):
		return self.write8(value>>8).write8(value)

	def write24(self, value):
		return self.write16(value>>8).write8(value)

	def write32(self, value):
		return self.write24(value>>8).write8(value)

	def writeBool(self, value):
		return self.write8(1 if value else 0)

	def writeString(self, string):
		if isinstance(string, str):
			string = string.encode()
		return self.write16(len(string)).writeBytes(string)

	def export(self, fp=0):
		if self.exported:
			return self.bytes

		m = Packet()
		size = len(self.buffer)
		if size<=0xff:
			m.write8(1).write8(size)
		elif size<=0xffff:
			m.write8(2).write16(size)
		elif size<=0xffffff:
			m.write8(3).write24(size)
		else:
			raise Exception('Packet too long')
		m.write8(fp)

		self.bytes = bytes(m.buffer + self.buffer)
		self.exported = True

		return self.bytes

	def xor_cipher(self, keys, fp):
		fp += 1
		ccc = self.readBytes(2)
		tmp = bytearray([(byte^keys[(fp+i)%20])&0xff for i, byte in enumerate(self.buffer[2:])])
		self.buffer = ccc+tmp
		return self

	def cipher(self, keys):
		if len(self.buffer)<2:
			raise Exception()
		while len(self.buffer)<10:
			self.write8(0)

		chunks = []
		ccc = self.read16()
		length = len(self.buffer)-2
		for i in range(length//4+(length%4>0)):
			chunks.append(self.read32())

		chunks = encode_chunks(chunks, len(chunks), keys)

		packet = Packet.new(ccc).write16(len(chunks))
		for chunk in chunks:
			packet.write32(chunk)

		self.buffer = packet.buffer
		return self

def encode_chunks(v, n, keys):
	DELTA = 0x9e3779b9
	def MX():
		return int(((z>>5)^(y<<2)) + ((y>>3)^(z<<4))^(sum^y) + (keys[(p & 3)^e]^z))

	y = v[0]
	sum = 0
	if n > 1:
		z = v[n - 1]
		q = int(6 + 52 / n)
	while q > 0:
		q -= 1
		sum = (sum + DELTA) & 0xffffffff
		e = ((sum >> 2) & 0xffffffff) & 3
		p = 0
		while p < n - 1:
			y = v[p + 1]
			z = v[p] = (v[p] + MX()) & 0xffffffff
			p += 1
		y = v[0]
		z = v[n - 1] = (v[n - 1] + MX()) & 0xffffffff
	return v


import re
import setuptools

with open('aiotfm/__version__.py') as f:
	match = re.search(r'__version__ = (["\'])([^"\']+)\1', f.read())
	version = match.group(2)

with open('requirements.txt') as f:
	requirements = f.readlines()
with open('README.md') as f:
	README = f.read()

setuptools.setup(
	name='aiotfm',
	version=version,
	packages=['aiotfm'],
	author='Athesdrake',
	description="An asynchronous event based client for Transformice.",
	long_description=README,
	long_description_content_type='text/markdown',
	url='https://github.com/Athesdrake/aiotfm',
	install_requires=requirements,
	python_requires='>=3.5.3',
	keywords=['TRANFORMICE', 'CLIENT', 'ASYNC', 'ATELIER801', 'EVENT'],
	classifiers=[
		"Development Status :: 2 - Pre-Alpha",
		"Framework :: AsyncIO",
		"Intended Audience :: Developers",
		"Programming Language :: Python :: 3.5",
		"Programming Language :: Python :: 3.6",
		"Topic :: Software Development :: Libraries",
	]
)


import aiohttp

async def get_keys(tfm_id, api_token):
	url = 'https://api.tocu.tk/get_transformice_keys.php?tfmid={}&token={}'.format(tfm_id, api_token)
	async with aiohttp.ClientSession() as session:
		async with session.get(url) as resp:
			data = await resp.json()
	# result = request.urlopen(Request(url, headers={"User-Agent": "Mozilla/5.0"})).read()
	# data = json.loads(result)

	if data['success']:
		if not data['internal_error']:
			class keys:
				version = data['version']
				connection = data['connection_key']
				auth = data['auth_key']
				packet = data['packet_keys']
				identification = data['identification_keys']
				msg = data['msg_keys']
			return keys
		else:
			if data['internal_error_step']==2:
				raise Exception('The game might be in maintenance mode.')
			else:
				raise Exception('An internal error occur: {}'.format(data['internal_error_step']))
	else:
		raise Exception("Can't get the keys. Error info: {}".format(data['error']))